/* Copyright (c) 2008 Markus Peloquin <markus@cs.wisc.edu>
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE. */

#ifndef FIELD_HPP
#define FIELD_HPP

#include <cassert>
#include <cmath>
#include <limits>
#include <list>
#include <memory>
#include <tr1/memory>
#include <vector>

#include "person.hpp"

// seemingly optimal for uniform distributions
const float PEOPLE_PER_SQUARE = 2.5f;

class Field {
public:
	// accepts a sequence of shared_ptr<Person>
	template <class Fwd>
	Field(Fwd start, Fwd end);

	void find_close();

private:
	struct Node {
		Node() : people() {}
		typedef
		    std::vector<std::tr1::shared_ptr<Person> >::iterator
		    iterator;
		typedef
		    std::vector<std::tr1::shared_ptr<Person> >::const_iterator
		    const_iterator;

		std::vector<std::tr1::shared_ptr<Person> > people;
	};

	struct Search_area {
		Search_area() : indices() {}
		typedef std::list<size_t>::iterator iterator;
		typedef std::list<size_t>::const_iterator const_iterator;

		std::list<size_t> indices;
	};

	size_t lat_to_row(double lat) const
	{
		return static_cast<size_t>(
		    (lat + 90.0f) * _rows / 180.01f
		);
	}
	size_t lon_to_col(double lon) const
	{
		return static_cast<size_t>(
		    (lon + 180.0f) * _cols / 360.01f
		);
	}
	size_t index_of(size_t row, size_t col) const
	{
		// column-major
		return row + col * _rows;
	}
	size_t index_of(const Person &person) const
	{
		return index_of(
		    lat_to_row(person.latitude()),
		    lon_to_col(person.longitude()));
	}

	// for each pair (p1,p2) in (n1,n2):
	//   p1 becomes neighbor of p2 if close
	void consider(Node &n1, const Node &n2);

	std::auto_ptr<Search_area> search_radius(size_t row, size_t col) const;

	std::vector<Node> _field;
	size_t _cols, _rows;
};

template <class Fwd>
Field::Field(Fwd start, Fwd end)
{
	size_t sz = std::distance(start, end);
	sz = std::distance(start, end);

	/* two desired properties:
	 * 1) r * c * x >= size, where x is the number of people per square
	 * 2) 2 * r = c, because the map is twice as wide as it is tall
	 * so: r >= sqrt(size / (2 * x)) */
	_rows = static_cast<size_t>(
	    ceil(sqrt(sz / (2 * PEOPLE_PER_SQUARE)))
	);

	_cols = 2 * _rows;

	_field.resize(_rows * _cols);

	// copy people into field
	for (Fwd i = start; i != end; ++i) {
		std::tr1::shared_ptr<Person> &p = *i;
		_field[index_of(*p)].people.push_back(p);
	}

	/*
	// compact all vectors
	for (std::vector<Node>::iterator i = _field.begin();
	    i != _field.end(); ++i) {
		std::vector<std::tr1::shared_ptr<Person> > t = i->people;
		i->people.swap(t);
	}
	*/
}

#endif
