/* Copyright (c) 2008 Markus Peloquin <markus@cs.wisc.edu>
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE. */

#ifndef PERSON_HPP
#define PERSON_HPP

#include <cassert>
#include <vector>
#include <tr1/memory>

const size_t CLOSE_FRIENDS = 3;

class Person;

// gives the square of the distance
inline double	dist(const Person &p1, const Person &p2);

struct Neighbor {
	std::tr1::shared_ptr<const Person> p;
	double d;
};

class Person {
public:
	Person(int id, double lat, double lon) :
	    _neighbors(), _id(id), _lat(lat), _lon(lon)
	{
		assert(lat >= -90.0f);  assert(lat <= 90.0f);
		assert(lon >= -180.0f); assert(lat <= 180.0f);
		_neighbors.reserve(CLOSE_FRIENDS + 1);
	}

	Person(const Person &p) :
	    _neighbors(p._neighbors), _id(p._id), _lat(p._lat), _lon(p._lon)
	{}

	Person &operator=(const Person &p)
	{
		Person copy = p;
		swap(copy);
		return *this;
	}

	void swap(Person &p)
	{
		std::swap(_neighbors, p._neighbors);
		std::swap(_id, p._id);
		std::swap(_lat, p._lat);
		std::swap(_lon, p._lon);
	}

	void add_if_close(std::tr1::shared_ptr<const Person> p, double dist);

	void add_if_close(std::tr1::shared_ptr<const Person> p)
	{
		add_if_close(p, dist(*this, *p));
	}

	// order of neighbors by decreasing dist
	const std::vector<Neighbor> &neighbors() const
	{ return _neighbors; }

	int id() const
	{ return _id; }

	double latitude() const
	{ return _lat; }

	double longitude() const
	{ return _lon; }

	bool operator==(const Person &p) const
	{ return _id == p._id; }

	bool operator!=(const Person &p) const
	{ return _id != p._id; }

private:
	// order of neighbors by decreasing dist
	std::vector<Neighbor> _neighbors;

	int _id;
	double _lat;
	double _lon;
};

inline double
dist(const Person &p1, const Person &p2)
{
#ifdef CORRECT_WAY
	// in the correct method, longitude wrap around; this is not what
	// facebook wants
	double lon = abs(p1.longitude() - p2.longitude());
	if (lon > 180) lon = 360 - lon;
#else
	double lon = p1.longitude() - p2.longitude();
#endif
	double lat = p1.latitude() - p2.latitude();
	return lat * lat + lon * lon;
}

#endif
